/*
fireprofile.c - FireProfile preload file
Copyright (C) 2000 Ian Gulliver

This program is free software; you can redistribute it and/or modify
it under the terms of version 2 of the GNU General Public License as
published by the Free Software Foundation.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <unistd.h>
#include <mysql/mysql.h>
#include <unistd.h>
#define __USE_GNU
#include <dlfcn.h>

static int (*real_mysql_real_query)(MYSQL *, const char *, unsigned int);
static MYSQL* (*real_mysql_real_connect)(MYSQL *, const char *, const char *, const char *, const char *, unsigned int, const char *, unsigned int clientflag);
static MYSQL_RES* (*real_mysql_store_result)(MYSQL *);
static MYSQL_RES* (*real_mysql_use_result)(MYSQL *);

static struct timeval starttime;
static double totalconnect = 0, totalquery = 0, totalresult = 0;

void _init() {
	fprintf(stderr,">>> libfireprofile starting...\n");
	gettimeofday(&starttime,NULL);
	real_mysql_real_query = dlsym(RTLD_NEXT,"mysql_real_query");
	if (real_mysql_real_query == NULL) {
		fprintf(stderr,">>>--- no libmysqlclient link found; exiting\n");
		exit(EXIT_FAILURE);
	}
	real_mysql_real_connect = dlsym(RTLD_NEXT,"mysql_real_connect");
	if (real_mysql_real_connect == NULL) {
		fprintf(stderr,">>>--- no libmysqlclient link found; exiting\n");
		exit(EXIT_FAILURE);
	}
	real_mysql_store_result = dlsym(RTLD_NEXT,"mysql_store_result");
	if (real_mysql_store_result == NULL) {
		fprintf(stderr,">>>--- no libmysqlclient link found; exiting\n");
		exit(EXIT_FAILURE);
	}
	real_mysql_use_result = dlsym(RTLD_NEXT,"mysql_use_result");
	if (real_mysql_use_result == NULL) {
		fprintf(stderr,">>>--- no libmysqlclient link found; exiting\n");
		exit(EXIT_FAILURE);
	}
}

MYSQL_RES* mysql_store_result(MYSQL *mysql) {
	MYSQL_RES *ret;
	struct timeval storestart, storeend;
	double storelength;

	fprintf(stderr,">>> mysql store_result\n");
	gettimeofday(&storestart,NULL);
	ret = real_mysql_store_result(mysql);
	gettimeofday(&storeend,NULL);
	storelength = (storeend.tv_sec - storestart.tv_sec) + ((storeend.tv_usec - storestart.tv_usec) / 1000000.0);
	totalresult += storelength;
	fprintf(stderr,">>>--- store_result time: %f\n",storelength);
	return ret;
}

MYSQL_RES* mysql_use_result(MYSQL *mysql) {
	MYSQL_RES *ret;
	struct timeval usestart, useend;
	double uselength;

	fprintf(stderr,">>> mysql use_result\n");
	gettimeofday(&usestart,NULL);
	ret = real_mysql_use_result(mysql);
	gettimeofday(&useend,NULL);
	uselength = (useend.tv_sec - usestart.tv_sec) + ((useend.tv_usec - usestart.tv_usec) / 1000000.0);
	totalresult += uselength;
	fprintf(stderr,">>>--- use_result time: %f\n",uselength);
	fprintf(stderr,">>>!!! using mysql_use_result() is almost always a bad idea, as the server may hold locks while you are processing the data.  use mysql_store_result() instead.\n");
	return ret;
}

MYSQL *mysql_real_connect(MYSQL *mysql, const char *host, const char *user, const char *passwd, const char *db, unsigned int port, const char *unix_socket, unsigned int clientflag) {
	MYSQL *ret;
	struct timeval connectstart, connectend;
	double connectlength;

	fprintf(stderr,">>> mysql connect: %s:%s@%s:%d %s\n",user,passwd,host,port,db);
	gettimeofday(&connectstart,NULL);
	ret = real_mysql_real_connect(mysql,host,user,passwd,db,port,unix_socket,clientflag);
	gettimeofday(&connectend,NULL);
	connectlength = (connectend.tv_sec - connectstart.tv_sec) + ((connectend.tv_usec - connectstart.tv_usec) / 1000000.0);
	totalconnect += connectlength;
	fprintf(stderr,">>>--- connect time: %f\n",connectlength);
	return ret;
}

#define PREFIX "EXPLAIN "
int mysql_real_query(MYSQL *mysql, const char *query, unsigned int length) {
	int ret;
	struct timeval querystart, queryend;
	double querylength;
	char *ourquery;
	MYSQL_RES *res;
	MYSQL_ROW row;

	fprintf(stderr,">>> mysql query: '%s'\n",query);
	ourquery = malloc(length + sizeof(PREFIX) - 1);
	if (ourquery == NULL) {
		perror("malloc");
		exit(EXIT_FAILURE);
	}
	strcpy(ourquery,PREFIX);
	memcpy(&ourquery[sizeof(PREFIX) - 1],query,length);
	ret = real_mysql_real_query(mysql,ourquery,length + sizeof(PREFIX) - 1);
	if (ret == 0) {
		res = mysql_store_result(mysql);
		if (res != NULL) {
			while ((row = mysql_fetch_row(res)) != NULL) {
				int i,n;
				MYSQL_FIELD *fields;
				n = mysql_num_fields(res);
				fields = mysql_fetch_fields(res);
				for (i = 0; i < n; i++) {
					fprintf(stderr,">>>--- %s: %s\n",fields[i].name,row[i]);
					if (strcmp(fields[i].name,"type") == 0 && strcmp(row[i],"ALL") == 0)
						fprintf(stderr,">>>!!! query type is ALL.  this is bad; you should create an index to match this query.\n");
				}
				fprintf(stderr,">>>---------\n");
			}
			mysql_free_result(res);
		}
	}
	gettimeofday(&querystart,NULL);
	ret = real_mysql_real_query(mysql,query,length);
	gettimeofday(&queryend,NULL);
	querylength = (queryend.tv_sec - querystart.tv_sec) + ((queryend.tv_usec - querystart.tv_usec) / 1000000.0);
	totalquery += querylength;
	fprintf(stderr,">>>--- query time: %f\n",querylength);
	return ret;
}

void _fini() {
	struct timeval curtime;
	struct rusage usage;
	double runtime, usertime, system, mysqltime;

	fprintf(stderr,">>> libfireprofile exiting...\n");
	gettimeofday(&curtime,NULL);
	getrusage(RUSAGE_SELF,&usage);
	runtime = (curtime.tv_sec - starttime.tv_sec) + ((curtime.tv_usec - starttime.tv_usec) / 1000000.0);
	usertime = usage.ru_utime.tv_sec + (usage.ru_utime.tv_usec / 1000000.0);
	system = usage.ru_stime.tv_sec + (usage.ru_stime.tv_usec / 1000000.0);
	mysqltime = totalconnect + totalquery + totalresult;
	fprintf(stderr,">>>--- total runtime:  %f\n",runtime);
	fprintf(stderr,">>>------ user time:   %f\n",usertime);
	fprintf(stderr,">>>------ system time: %f\n",system);
	fprintf(stderr,">>>------ mysql time:  %f\n",mysqltime);
	fprintf(stderr,">>>--------- connect time: %f\n",totalconnect);
	fprintf(stderr,">>>--------- query time:   %f\n",totalquery);
	fprintf(stderr,">>>--------- result time:  %f\n",totalresult);
	fprintf(stderr,">>>--- libfireprofile done.\n");
}
